import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.io.*;

public class ImageForFFT {

	protected int readInt(DataInputStream in) throws IOException {
		int a, b, c, d;
		a = in.readUnsignedByte();
		b = in.readUnsignedByte();
		c = in.readUnsignedByte();
		d = in.readUnsignedByte();
		return (d << 24) + (c << 16) + (b << 8) + a;
	}

	protected int readShort(DataInputStream in) throws IOException {
		int a, b;
		a = in.readUnsignedByte();
		b = in.readUnsignedByte();
		return (b << 8) + a;
	}

	protected void writeInt(DataOutputStream out, int n) throws IOException {
		int a, b, c, d;
		a = n & 0xFF;
		b = (n >> 8) & 0xFF;
		c = (n >> 16) & 0xFF;
		d = (n >> 24) & 0xFF;
		out.writeByte(a);
		out.writeByte(b);
		out.writeByte(c);
		out.writeByte(d);
	}

	protected void writeShort(DataOutputStream out, int n) throws IOException {
		int a, b;
		a = n & 0xFF;
		b = (n >> 8) & 0xFF;
		out.writeByte(a);
		out.writeByte(b);
	}

	public ImageForFFT(int RedArray[][], int GreenArray[][], int BlueArray[][], int width, int height) {
		this.RedArray = RedArray;
		this.GreenArray = GreenArray;
		this.BlueArray = BlueArray;
		this.width = width;
		this.height = height;
	}

	public ImageForFFT(String fn) {
		DataInputStream in;

		try {
			in = new DataInputStream(new BufferedInputStream(new FileInputStream(fn)));
		} catch(FileNotFoundException nfe) {
			System.out.println("Fichier \"" + fn + "\" not found !");
//			System.exit(1);
			return;
		}

		try {
			if(readShort(in) == 0x4D42) {
				readInt(in);		// File size
				readShort(in);
				readShort(in);

				int OffsetBits = readInt(in);
				readInt(in);
				width = readInt(in);
				height= readInt(in);
				in.readShort();		// planes
				int BitsCount = readShort(in);		// bits count
				if(BitsCount == 24) {
					int compression = readInt(in);
					if(compression == 0) {
						in.skipBytes(4+4+4+4+4);
						BlueArray = new int[width][height];
						GreenArray = new int[width][height];
						RedArray = new int[width][height];
						for(int y = 0 ; y < height ; y++) {
							for(int x = 0 ; x < width ; x++) {
								BlueArray[x][y] = in.readUnsignedByte();
								GreenArray[x][y] = in.readUnsignedByte();
								RedArray[x][y] = in.readUnsignedByte();
							}
//							in.readUnsignedByte();
//							in.readUnsignedByte();
						}
					}
				}
			} else {
				System.out.println("Unknown file format !");
			}
			if(in != null) in.close();
		} catch(IOException ioe) {
			System.out.println("Error reading the file.");
//			System.exit(1);
			return;
		}
	}

	public void save(String fn) {
		DataOutputStream out;

		try {
			out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(fn)));
		} catch(FileNotFoundException nfe) {
			System.out.println("Fichier \"" + fn + "\" not created !");
//			System.exit(1);
			return;
		}

		try {
			writeShort(out, 0x4D42);
			writeInt(out, 0x36 + width * height * 3);		// File size
			writeShort(out, 0);
			writeShort(out, 0);

			writeInt(out, 0x36);					// offset bits
			writeInt(out, 0x28);
			writeInt(out, width);
			writeInt(out, height);
			writeShort(out, 1);		// planes
			writeShort(out, 24);		// bits count
			writeInt(out, 0);	// compression

			writeInt(out, 0);
			writeInt(out, 0);
			writeInt(out, 0);
			writeInt(out, 0);
			writeInt(out, 0);

			for(int y = 0 ; y < height ; y++) {
				for(int x = 0 ; x < width ; x++) {
					out.writeByte(BlueArray[x][y]);
					out.writeByte(GreenArray[x][y]);
					out.writeByte(RedArray[x][y]);
				}
			}
			if(out != null) out.close();
		} catch(IOException ioe) {
			System.out.println("Error writing the file.");
//			System.exit(1);
			return;
		}
	}

	protected int getRGB(int x, int y) {
		return (RedArray[x][y] << 16) + (GreenArray[x][y] << 8) + BlueArray[x][y];
	}

	public Image getImage() {
		BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);

		for(int y = 0 ; y < height ; y++)
			for(int x = 0 ; x < width ; x++)
				image.setRGB(x, (height - 1) - y, getRGB(x, y));
		return image;
	}

	public ImageForFFT applyFilter(int filter, int cutoff1, int cutoff2, ImageForFFT img, int dest) {
		ImageForFFT ret = null;
		COMPLEX redgrid[][] = computeGrid(RedArray);
		COMPLEX greengrid[][] = computeGrid(GreenArray);
		COMPLEX bluegrid[][] = computeGrid(BlueArray);

		FFT2D(redgrid, width, height, 1);
		FFT2D(greengrid, width, height, 1);
		FFT2D(bluegrid, width, height, 1);

		int redfilter[][] = null, greenfilter[][] = null, bluefilter[][] = null;
		if(img != null) {
			redfilter = img.getRed();
			greenfilter = img.getGreen();
			bluefilter = img.getBlue();
		}

		int r, r1, r2, r3, r4;
		for (int i=0;i<width;i++) {
			for (int j=0;j<height;j++) {

				r1 = i*i + j*j;
				r2 = (i-width)*(i-width) + j*j;
				r3 = i*i + (j-height)*(j-height);
				r4 = (i-width)*(i-width) + (j-height)*(j-height);
				r = Math.min(r1,r2);
				r = Math.min(r,r3);
				r = Math.min(r,r4);

				switch (filter) {
					case 0:		/* No filtering */
						break;
					case 1:		/* Low pass */
						if (r > cutoff1*cutoff1) {
							redgrid[i][j].real = 0;
							redgrid[i][j].imag = 0;
							greengrid[i][j].real = 0;
							greengrid[i][j].imag = 0;
							bluegrid[i][j].real = 0;
							bluegrid[i][j].imag = 0;
						}
						break;
					case 2:		/* Band pass */
						if (r < cutoff1*cutoff1 || r > cutoff2*cutoff2) {
							redgrid[i][j].real = 0;
							redgrid[i][j].imag = 0;
							greengrid[i][j].real = 0;
							greengrid[i][j].imag = 0;
							bluegrid[i][j].real = 0;
							bluegrid[i][j].imag = 0;
						}
						break;
					case 3:     /* High pass */
						if (r < cutoff1*cutoff1) {
							redgrid[i][j].real = 0;
							redgrid[i][j].imag = 0;
							greengrid[i][j].real = 0;
							greengrid[i][j].imag = 0;
							bluegrid[i][j].real = 0;
							bluegrid[i][j].imag = 0;
						}
						break;
					case 4:
						if (r > 0) {
/*							redgrid[i][j].real = 0;
							redgrid[i][j].imag = 0;
							greengrid[i][j].real = 0;
							greengrid[i][j].imag = 0;
							bluegrid[i][j].real = 0;
							bluegrid[i][j].imag = 0;*/
							redgrid[i][j].real *= Math.pow(1/r,cutoff1);
							redgrid[i][j].imag *= Math.pow(1/r,cutoff1);
							greengrid[i][j].real *= Math.pow(1/r,cutoff1);
							greengrid[i][j].imag *= Math.pow(1/r,cutoff1);
							bluegrid[i][j].real *= Math.pow(1/r,cutoff1);
							bluegrid[i][j].imag *= Math.pow(1/r,cutoff1);
						}
						break;
					case 5:
						if(img != null) {

					int ii, jj;
					if (i <= width/2) {
						if (j <= height/2) {
							ii = i+width/2-1;
							jj = j+height/2-1;
						} else {
							ii = i+width/2-1;
							jj = j-height/2-1;
						}
					} else {
						if (j <= height/2) {
							ii = i-width/2-1;
							jj = j+height/2-1;
						} else {
							ii = i-width/2-1;
							jj = j-height/2-1;
						}
					}

							redgrid[i][j].real = (redgrid[i][j].real * (((double)redfilter[ii][jj]) / 255));
							redgrid[i][j].imag = (redgrid[i][j].imag * (((double)redfilter[ii][jj]) / 255));
							greengrid[i][j].real = (greengrid[i][j].real * (((double)greenfilter[ii][jj]) / 255));
							greengrid[i][j].imag = (greengrid[i][j].imag * (((double)greenfilter[ii][jj]) / 255));
							bluegrid[i][j].real = (bluegrid[i][j].real * (((double)bluefilter[ii][jj]) / 255));
							bluegrid[i][j].imag = (bluegrid[i][j].imag * (((double)bluefilter[ii][jj]) / 255));
						} else {
							System.out.println("No filter specified");
//							System.exit(-1);
							return null;
						}
						break;
					default:
						System.out.println("No such filter defined");
//						System.exit(-1);
						break;
				}
			} /* j */
		} /* i */

		int RedArray2[][];
		int GreenArray2[][];
		int BlueArray2[][];
		if(dest == 1) {
			FFT2D(redgrid, width, height, -1);
			FFT2D(greengrid, width, height, -1);
			FFT2D(bluegrid, width, height, -1);

			double s = computeScale(redgrid, 0);
			s = computeScale(greengrid, s);
			s = computeScale(bluegrid, s);

			RedArray2 = computeFromGrid(redgrid, 'e', s);
			GreenArray2 = computeFromGrid(greengrid, 'e', s);
			BlueArray2 = computeFromGrid(bluegrid, 'e', s);
		} else {
			double s = computeScale(redgrid, 0);
			s = computeScale(greengrid, s);
			s = computeScale(bluegrid, s);

			RedArray2 = computeFromGrid(redgrid, 'f', s);
			GreenArray2 = computeFromGrid(greengrid, 'f', s);
			BlueArray2 = computeFromGrid(bluegrid, 'f', s);
		}
		ret = new ImageForFFT(RedArray2, GreenArray2, BlueArray2, width, height);

		return ret;
	}

	public ImageForFFT applyWindow(ImageForFFT window) {
		ImageForFFT r = null;
		
		int RedArray2[][] = computeWindow(getRed(), window.getRed());
		int GreenArray2[][] = computeWindow(getGreen(), window.getGreen());
		int BlueArray2[][] = computeWindow(getBlue(), window.getBlue());
		r = new ImageForFFT(RedArray2, GreenArray2, BlueArray2, width, height);

		return r;
	}

	protected int[][] computeWindow(int Array[][], int window[][]) {
		int Array2[][] = new int[width][height];

		for (int i = 0 ; i < width ; i++)
			for (int j = 0 ; j < height ; j++)
				Array2[i][j] = (int)(Array[i][j] * ((double)window[i][j]) / 255);

		return Array2;
	}

	public int[][] getRed() {
		return RedArray;
	}

	public int[][] getGreen() {
		return GreenArray;
	}

	public int[][] getBlue() {
		return BlueArray;
	}

	public ImageForFFT applyFFT(int dir) {
		ImageForFFT r = null;
		COMPLEX gridred[][] = computeGrid(RedArray);
		COMPLEX gridgreen[][] = computeGrid(GreenArray);
		COMPLEX gridblue[][] = computeGrid(BlueArray);

		FFT2D(gridred, width, height, dir);
		FFT2D(gridgreen, width, height, dir);
		FFT2D(gridblue, width, height, dir);

		double s = computeScale(gridred, 0);
		s = computeScale(gridgreen, s);
		s = computeScale(gridblue, s);
		int RedArray2[][] = computeFromGrid(gridred, 'f', s);
		int GreenArray2[][] = computeFromGrid(gridgreen, 'f', s);
		int BlueArray2[][] = computeFromGrid(gridblue, 'f', s);
		r = new ImageForFFT(RedArray2, GreenArray2, BlueArray2, width, height);
		return r;
	}

	protected double computeScale(COMPLEX[][] c, double scale) {
		double mag;
		for (int i = 0 ; i < width ; i++) {
			for (int j = 0 ; j < height ; j++) {
				mag = Math.sqrt(c[i][j].real*c[i][j].real+c[i][j].imag*c[i][j].imag);
				scale = Math.max(mag,scale);
			}
		}
		return scale;
	}

	protected int[][] computeFromGrid(COMPLEX[][] c, char type, double scale) {
		double mag, rmin = 1e32, rmax = -1e32;
		int Array[][] = new int[width][height];
		int ii, jj;

		/* Determine the maximum magnitude */
		for (int i = 0 ; i < width ; i++) {
			for (int j = 0 ; j < height ; j++) {
//				mag = Math.sqrt(c[i][j].real*c[i][j].real+c[i][j].imag*c[i][j].imag);
//				scale = Math.max(mag,scale);
				rmin = Math.min(rmin,c[i][j].real);
				rmax = Math.max(rmax,c[i][j].real);
			}
		}
		if (rmin >= rmax) {
			rmin -= 0.01;
			rmax += 0.01;
		}

		/* Form the greyscale image */
		for (int i = 0 ; i < width ; i++) {
			for (int j = 0 ; j < height ; j++) {
				if (type == 'f') {
					mag = Math.sqrt(c[i][j].real*c[i][j].real+c[i][j].imag*c[i][j].imag);
					if (i <= width/2) {
						if (j <= height/2) {
							ii = i+width/2-1;
							jj = j+height/2-1;
						} else {
							ii = i+width/2-1;
							jj = j-height/2-1;
						}
					} else {
						if (j <= height/2) {
							ii = i-width/2-1;
							jj = j+height/2-1;
						} else {
							ii = i-width/2-1;
							jj = j-height/2-1;
						}
					}
//					Array[ii][jj] = 255 * (int)(0.25 * width * mag / scale);
					Array[ii][jj] = (int)(Math.min(3000 * mag / scale, 255));
				} else {
					ii = i;
					jj = j;
					Array[ii][jj] = (int)(Math.min(255 * (c[ii][jj].real - rmin) / (rmax - rmin), 255));
				} 
			}
		}

		return Array;
	}

	protected COMPLEX[][] computeGrid(int Array[][]) {
		double mean = 0;
		COMPLEX grid[][] = new COMPLEX[width][height];

		for(int i = 0 ; i < width ; i++)
			for(int j = 0 ; j < height ; j++) {
				grid[i][j] = new COMPLEX(Array[i][j], 0);
				mean += Array[i][j];
			}

		mean /= width * height;

		for(int i = 0 ; i < width ; i++)
			for(int j = 0 ; j < height ; j++)
				grid[i][j].real -= mean;

		return grid;
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}

	/*-------------------------------------------------------------------------
	   This computes an in-place complex-to-complex FFT
	   x and y are the real and imaginary arrays of 2^m points.
	   dir =  1 gives forward transform
	   dir = -1 gives reverse transform

	     Formula: forward
	                  N-1
	                  ---
	              1   \          - j k 2 pi n / N
	      X(n) = ---   >   x(k) e                    = forward transform
	              N   /                                n=0..N-1
	                  ---
	                  k=0

	      Formula: reverse
	                  N-1
	                  ---
	                  \          j k 2 pi n / N
	      X(n) =       >   x(k) e                    = forward transform
	                  /                                n=0..N-1
	                  ---
	                  k=0
	*/
	protected void FFT(int dir, int m, double x[], double y[])
	{
		int nn,i,i1,j,k,i2,l,l1,l2;
		double c1,c2,tx,ty,t1,t2,u1,u2,z;

		/* Calculate the number of points */
		nn = 1;
		for (i=0;i<m;i++)
			nn *= 2;

		/* Do the bit reversal */
		i2 = nn >> 1;
		j = 0;
		for (i=0;i<nn-1;i++) {
			if (i < j) {
				tx = x[i];
				ty = y[i];
				x[i] = x[j];
				y[i] = y[j];
				x[j] = tx;
				y[j] = ty;
			}
			k = i2;
			while (k <= j) {
				j -= k;
				k >>= 1;
			}
			j += k;
		}

		/* Compute the FFT */
		c1 = -1.0;
		c2 = 0.0;
		l2 = 1;
		for (l=0;l<m;l++) {
			l1 = l2;
			l2 <<= 1;
			u1 = 1.0;
			u2 = 0.0;
			for (j=0;j<l1;j++) {
				for (i=j;i<nn;i+=l2) {
					i1 = i + l1;
					t1 = u1 * x[i1] - u2 * y[i1];
					t2 = u1 * y[i1] + u2 * x[i1];
					x[i1] = x[i] - t1;
					y[i1] = y[i] - t2;
					x[i] += t1;
					y[i] += t2;
				}
				z =  u1 * c1 - u2 * c2;
				u2 = u1 * c2 + u2 * c1;
				u1 = z;
			}
			c2 = Math.sqrt((1.0 - c1) / 2.0);
			if (dir == 1)
				c2 = -c2;
			c1 = Math.sqrt((1.0 + c1) / 2.0);
		}

		/* Scaling for forward transform */
		if (dir == 1) {
			for (i=0;i<nn;i++) {
				x[i] /= (double)nn;
				y[i] /= (double)nn;
			}
		}

	}

	protected static int PowerOf2(int n) {
		int r = 1;
		for(int i = 0 ; i < n ; i++)
			r *= 2;
		return r;
	}

	protected static int ExponentOf2(int n) {
		return (int)(Math.log(n)/Math.log(2));
	}

	/*-------------------------------------------------------------------------
	   Perform a 2D FFT inplace given a complex 2D array
	   The direction dir, 1 for forward, -1 for reverse
	   The size of the array (nx,ny)
	   Return false if there are memory problems or
	      the dimensions are not powers of 2
	*/
	protected boolean FFT2D(COMPLEX c[][],int nx,int ny,int dir)
	{
		int i,j;
		int m,twopm;
		double real[],imag[];

		/* Transform the rows */
		real = new double[nx];
		imag = new double[nx];
		m = ExponentOf2(nx);
		twopm = PowerOf2(m);
		if ((m == 0) || (twopm != nx))
			return false;
		
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				real[i] = c[i][j].real;
				imag[i] = c[i][j].imag;
			}
			FFT(dir,m,real,imag);
			for (i=0;i<nx;i++) {
				c[i][j].real = real[i];
				c[i][j].imag = imag[i];
			}
		}

		/* Transform the columns */
		real = new double[ny];
		imag = new double[ny];
		m = ExponentOf2(ny);
		twopm = PowerOf2(m);
		if ((m == 0) || (twopm != ny))
			return false;
		for (i=0;i<nx;i++) {
			for (j=0;j<ny;j++) {
				real[j] = c[i][j].real;
				imag[j] = c[i][j].imag;
			}
			FFT(dir,m,real,imag);
			for (j=0;j<ny;j++) {
				c[i][j].real = real[j];
				c[i][j].imag = imag[j];
			}
		}

		return true;
	}

	public static void main(String args[]) {
		JFrame f = new JFrame();
		Container c = f.getContentPane();

		PanelImage p = new PanelImage();
		c.add(p);

		PanelImage p2 = new PanelImage();
		c.add(p2);

		ImageForFFT ifft = new ImageForFFT("square1.bmp");
		p.setImage(ifft.getImage(), ifft.getWidth(), ifft.getHeight());
		ImageForFFT idone = ifft.applyFFT(1);
		p2.setImage(idone.getImage(), idone.getWidth(), idone.getHeight());

		f.show();
	}

	protected class COMPLEX {
		public COMPLEX() {
			real = 0;
			imag = 0;
		}

		public COMPLEX(double real, double imag) {
			this.real = real;
			this.imag = imag;
		}

		public double real;
		public double imag;
	}

	protected int width, height, GreenArray[][], RedArray[][], BlueArray[][];
}
